-- $Rev: 133 $

ByeHakmud = {}
local BH = ByeHakmud

BH.CurrentPassengers = {}

--- Bye Hakmud Engine
function BH:GetEventFrame()
  return self.eventFrame
end

function BH:SetEventFrame(frame)
  self.eventFrame = frame
end

function BH:SetListen4Event(enabled, eventName)
    if (enabled) then
	if (not self:GetEventFrame():IsEventRegistered(eventName)) then
	    self:GetEventFrame():RegisterEvent(eventName); 
	end
    else 
	if (self:GetEventFrame():IsEventRegistered(eventName)) then
	    self:GetEventFrame():UnregisterEvent(eventName); 
	end
    end
end

function BH:SetListen4Passenger(enabled)
    self:SetListen4Event(enabled, "VEHICLE_PASSENGERS_CHANGED")
    self:SetListen4Event(enabled, "PLAYER_GAINS_VEHICLE_DATA")
end

function BH:AddTargetAsPermission(permission, presetName)
    local targetName = GetUnitName("target", false)
    if targetName then
	self:AddPlayerPermission(targetName, permission, presetName)
    end
    return targetName
end

function BH:IsPartyAllowsEject(presetName)
	local retVal = true;
	if self.Vars:IsOnlyEjectWhenInParty(presetName) then
	    retVal = UnitInParty("player")
	end
	return retVal;
end

function BH:GetSeatsToEject4Mount(creatureID, presetName)
    local retVal = nil;
    
    if self.Vars:GetMountIDToSeat(presetName) then
	retVal = self.Vars:GetMountIDToSeat(presetName)[creatureID]
    end
    
    if retVal == nil then
	retVal = BH_MountIDToSeatDefaultList[creatureID]
    end
    
    return retVal;
end

function BH:SetSeatsToEject4Mount(creatureID, seatsToEject)
    if self.Vars:GetMountIDToSeat(self.presetName) then 
	self.Vars:GetMountIDToSeat(self.presetName)[creatureID] = seatsToEject
    end
end

function BH:IsFriend(occupantName)
    local retVal = false
    for friendIdx = 1, GetNumFriends() do
	local friendName = GetFriendInfo(friendIdx)
	if friendName == occupantName then
	    retVal = true
	    break
	end
    end
    return retVal
end

function BH:CreateNotifyMessage(occupantName, creatureSpellID, presetName)
    local retVal = string.gsub(
	self.Vars:GetNotifyMsg(presetName),
	"%[passenger%]",
	tostring(occupantName)
    )
    
    local link, tradelink = GetSpellLink(creatureSpellID)
    retVal = string.gsub(
	retVal,
	"%[mount%]",
	tostring(link)
    )
    return retVal
end

function BH_MountInfoIter()
    local i = 1
    local mountCount = C_MountJournal.GetNumMounts()
    return function()
	local mountInfo
	if i < mountCount then
	    mountInfo = {}
	    mountInfo.creatureName, 
	    mountInfo.creatureSpellID, 
	    mountInfo.icon, 
	    mountInfo.active, 
	    mountInfo.isUsable, 
	    mountInfo.sourceType, 
	    mountInfo.isFavorite, 
	    mountInfo.isFactionSpecific, 
	    mountInfo.faction, 
	    mountInfo.hideOnChar, 
	    mountInfo.isCollected = C_MountJournal.GetMountInfoByID(i)
	    
	    i = i + 1
	end
	return mountInfo
    end
end

function BH_MountInfoByPlayerFactionIter()
    local playerFaction = (UnitFactionGroup("player") == "Alliance") and 1 or 0
    local baseMountInfoIter = BH_MountInfoIter()
    return function()
	local mountInfo
	repeat
	    mountInfo = baseMountInfoIter()
	until ((mountInfo == nil) or (mountInfo.isFactionSpecific == false) or (mountInfo.faction == playerFaction))
	return mountInfo
    end
end

local BH_DefaultSeatsToEject = {}
function BH:EjectHakmud(presetName)
    local creatureSpellID
    local function GetSeatsToEject()
	local function CreateDefaultSeatsToEject()
	    for seatIdx = 1, UnitVehicleSeatCount("player") do
		BH_DefaultSeatsToEject[seatIdx] = seatIdx
		BH_DefaultSeatsToEject[seatIdx + 1] = nil
	    end	
	    return BH_DefaultSeatsToEject
	end 
	
	local retVal = nil;
	for mountInfo in BH_MountInfoIter() do
	    if mountInfo.active then
		creatureSpellID = mountInfo.creatureSpellID
		retVal = self:GetSeatsToEject4Mount(creatureSpellID, presetName);
		
		if retVal == nil then
		    retVal = CreateDefaultSeatsToEject()
		end
		
		break;
	    end
	end
	
	-- druid stag travel form has no creatureID but a seat
	local class, classFileName = UnitClass("player")
	if classFileName == "DRUID" and GetShapeshiftFormID() == 3 then 
	    retVal = CreateDefaultSeatsToEject()
	    creatureSpellID = 783
	end
	
	return retVal;
    end

    local keepHakmud = IsShiftKeyDown();
    if (not self.Vars:IsHakmudKeptOnShift(presetName)) then
	keepHakmud = not keepHakmud;
    end

    if ((not keepHakmud)
	and self:IsPartyAllowsEject(presetName))
    then 
	local seatsToEject = GetSeatsToEject();
	if seatsToEject then
	    for i, seat in ipairs(seatsToEject) do
		if CanEjectPassengerFromSeat(seat) then
		    local controlType, occupantName, occupantRealm, canEject, canSwitchSeats = UnitVehicleSeatInfo("player", seat)
		    if self:IsEjectable(occupantName, presetName) then
			EjectPassengerFromSeat(seat)
			self:PostMessage(self:CreateNotifyMessage(occupantName, creatureSpellID, presetName), self.Vars:GetChannelToNotify(presetName))
		    end
		end
	    end
	end
    end
end

--- Player Permissions
local BH_DefaultPlayerPermissions = {
    "target",
    "focus",
    "friend",
    "guild",
    "group"
}
function BH:GetDefaultPlayerPermissions()
    return BH_DefaultPlayerPermissions
end


local BH_LastPrePlayerRule = "focus"

function BH_GetUnitName(unit)
    local name = GetUnitName(unit, false);
    if (name) then
        local separatorIdx = string.find(name, " ");
        if (separatorIdx and (separatorIdx > 0)) then
            name = string.sub(name, 0, separatorIdx -1);
        end
    end
    return name;
end

local BH_PermissionRules = {
    target = function(occupantName, presetName)
	return BH_GetUnitName("target") == occupantName;
    end,
    focus = function (occupantName, presetName)
	return BH_GetUnitName("focus") == occupantName 
    end,
    friend = function (occupantName, presetName)
	return BH:IsFriend(occupantName)
    end,
    guild = function (occupantName, presetName)
	return UnitIsInMyGuild(occupantName)
    end,
    group = function (occupantName, presetName)
	return true
    end,
}

function BH:IsEjectable(occupantName, presetName)
    local ejectable = false
    if presetName and UnitIsPlayer(occupantName) then
	for ruleName, permission in self.PairsByKeys(self:GetPlayerPermissions(presetName)) do
	    if permission ~= "DontCheck" then
		rule = BH_PermissionRules[ruleName]
		if (rule == nil and ruleName == occupantName) or 
		   (rule and rule(occupantName, presetName)) 
		then
		    ejectable = permission == "Deny"
		    break;
		end
	    end
	end
    else
	ejectable = true
    end
    
    return ejectable
end 

local BH_GetPlayerPermissionsRecursionBlock = false
function BH:GetPlayerPermissions(presetName)
    local retVal = self.Vars:GetPlayerPermissions(presetName)
    
    if retVal then 
	local k,v = next(retVal)
	if k == nil and BH_GetPlayerPermissionsRecursionBlock == false then
	    BH_GetPlayerPermissionsRecursionBlock = true
	    for i, name in pairs(BH_DefaultPlayerPermissions) do
		self:AddPlayerPermission(name, "DontCheck", presetName)
	    end
	    BH_GetPlayerPermissionsRecursionBlock = false
	end
    end
    
    return retVal
end

function BH.PairsByKeys (t)
    local a = {}
    for n in pairs(t) do 
	table.insert(a, n) 
    end
    table.sort(a, BH.PlayerPermissionsSort)
    
    local i = 0      -- iterator variable
    local iter = function ()   -- iterator function
	i = i + 1
	if a[i] == nil then 
	    return nil
	else 
	    return a[i], t[a[i]]
	end
    end
    return iter
end

function BH.PlayerPermissionsSort(a, b)
    local retVal
    
    local defaultIdxA, defaultIdxB, lastPrePlayerIdx
    for i, defaultName in ipairs(BH_DefaultPlayerPermissions) do
	if defaultName == a then defaultIdxA = i end
	if defaultName == b then defaultIdxB = i end
	if defaultName == BH_LastPrePlayerRule then lastPrePlayerIdx = i end
    end

    if defaultIdxA then
	if defaultIdxB then
	    retVal = defaultIdxA < defaultIdxB
	else
	    retVal = defaultIdxA <= lastPrePlayerIdx 
	end
    elseif defaultIdxB then
	retVal = defaultIdxB > lastPrePlayerIdx
    else 
	retVal = string.upper(a) < string.upper(b)
    end
    
    return retVal
end

function BH:GetPlayerPermission(player, presetName)
    local retVal = self:GetPlayerPermissions(presetName)[player] 
    if retVal == nil then
	retVal = "DontCheck"
    end
    return retVal
end

function BH:AddPlayerPermission(playerName, permission, presetName)
    local playerPermissions = self:GetPlayerPermissions(presetName)
    if playerName and playerPermissions then
	local trimmedName = playerName:gsub("^%s*(.-)%s*$", "%1")
	
	playerPermissions[trimmedName] = permission
	table.sort(playerPermissions, self.PlayerPermissionsSort)
    end
end

function BH:DelPlayerPermission(name, presetName)
    local reallyDelete = true
    for i, defaultPermissionName in ipairs(BH_DefaultPlayerPermissions) do
	if name == defaultPermissionName then
	    reallyDelete = false
	    break
	end
    end
    
    if reallyDelete then
	local playerPermissions = self:GetPlayerPermissions(presetName)
	playerPermissions[name] = nil
	table.sort(playerPermissions, self.PlayerPermissionsSort)
    end
end

local BH_InstanceTypes = {"none", "scenario", "party", "raid", "arena", "pvp" }
function BH:GetInstanceTypes()
  return BH_InstanceTypes
end

function BH:GetPresets4Situation(situation)
    local function IsPresetMatch(testPresetName)
      local regionMatch = false
      if situation.instanceType == "scenario" then
	  regionMatch = self.Vars:IsUsedInScenario(testPresetName)
      elseif situation.instanceType == "party" then
	  regionMatch = self.Vars:IsUsedInDungeon(testPresetName)
      elseif situation.instanceType == "raid" then
	  regionMatch = self.Vars:IsUsedInRaid(testPresetName)
      elseif situation.instanceType == "arena" then
	  regionMatch = self.Vars:IsUsedInArena(testPresetName)
      elseif situation.instanceType == "pvp" then
	  regionMatch = self.Vars:IsUsedInBattlegrounds(testPresetName)
      else
	  regionMatch = self.Vars:IsUsedInWorld(testPresetName)
      end
      
      local groupMatch = false
      if situation.numGuildPresent > 0 then
	  if situation.numGuildPresent == situation.numGroupMembers then
	      groupMatch = self.Vars:IsUsedInGuildParty(testPresetName)
	  else
	      groupMatch = self.Vars:IsUsedInMixedParty(testPresetName)
	  end
      else
	  groupMatch = self.Vars:IsUsedInRandomParty(testPresetName)
      end
      
      local pvpMatch = false
      if situation.isPvP then
	  pvpMatch = self.Vars:IsUsedInPvP(testPresetName)
      else
	  pvpMatch = self.Vars:IsUsedInPvE(testPresetName)
      end
      
      return regionMatch and groupMatch and pvpMatch
    end
    
    if self.situationPresets then
	for i = 1, #self.situationPresets do
	    self.situationPresets[i] = nil
	end
    else
	self.situationPresets = {}
    end
    
    local situationPresetIdx = 1
    for testPresetName, preset in pairs(BH.Vars:GetPresets()) do
	if IsPresetMatch(testPresetName) then 
	    self.situationPresets[situationPresetIdx] = testPresetName
	    situationPresetIdx = situationPresetIdx + 1
	end
    end
    
    return self.situationPresets
end
    
function BH:DetectSituation()
    if self.situation == nil then
	self.situation = {}
    end
    
    local inInstance, instanceType = IsInInstance()
    local inScenario = IsInScenarioGroup("player")
    
    if inScenario then
	self.situation.instanceType = "scenario"
    else 
	self.situation.instanceType = instanceType
    end
    
    self.situation.isPvP = UnitIsPVP("player")

    
    local inGroup, numGuildPresent, numGuildRequired, xpMultiplier = InGuildParty()
    self.situation.numGroupMembers = GetNumGroupMembers()
    self.situation.numGuildPresent = numGuildPresent
    
    local i, retVal = next(self:GetPresets4Situation(self.situation))
    
--~     print("InstanceType: "..tostring(instanceType).."("..tostring(self.situation.instanceType)..")"
--~       .." Scenario: "..tostring(inScenario)
--~       .." GuildGroup: "..tostring(numGuildPresent).."/"..tostring(numGuildRequired).."/"..tostring(self.situation.numGroupMembers)
--~       .." PvP: "..tostring(self.situation.isPvP)
--~       .." situation: "..tostring(retVal)
--~     )
    return retVal
end

--- Bye Hakmud Engines Event Handlers

function BH:OnAddonLoaded(addonName)
	if (addonName == "ByeHakmud") then
		self:GetEventFrame():UnregisterEvent("ADDON_LOADED");
		
		local ititialized = ByeHakmud_SavedVariables
		ByeHakmud_SavedVariables = ByeHakmud_SavedVariables or {}
		ByeHakmud_SavedVariablesPerCharacter = ByeHakmud_SavedVariablesPerCharacter or {}
		self.Vars:SetSavedVariables(ByeHakmud_SavedVariables)
		self.Vars:SetSavedVariablesPerCharacter(ByeHakmud_SavedVariablesPerCharacter)
		
		if ititialized == false then
		  self.Vars:CreatePreset("default")
		end
		
		-- for howto save missing translations see localization.core.lua
		if ByeHakmud_GetMissingTranslations() then 
		  ByeHakmud_SavedVariables.missingTranslations = ByeHakmud_GetMissingTranslations()
		else 
		  ByeHakmud_SavedVariables.missingTranslations = nil
		end
		self.Vars.loaded = true;
		
		self:SetListen4Passenger(true)
	end
end

function BH:OnPlayerGainsVehicleData()
    for i = 1, #self.CurrentPassengers do
	self.CurrentPassengers[i] = nil
    end
end

function BH:OnVehiclePassengersChanged(presetName)
    local changeDetected
    for seatIdx = 1, UnitVehicleSeatCount("player") do
	local 
	controlType, 
	occupantName, 
	occupantRealm, 
	canEject, 
	canSwitchSeats = UnitVehicleSeatInfo("player", seatIdx)
	
	-- mark empty seats as false instead of nil to avoid gaps in the BH.CurrentPassengers
	if occupantName == nil then 
	    occupantName = false 
	end
	
	if occupantName and 
	    ((self.CurrentPassengers[seatIdx] ~= occupantName) or
	    UnitIsPlayer(occupantName)) then
	    changeDetected = true
	end
	
	self.CurrentPassengers[seatIdx] = occupantName
    end
    
    if changeDetected then
	self:EjectHakmud(presetName)
    end
end

function BH:OnEvent(source, event, arg1, ...)
    if event == "ADDON_LOADED" then 
      self:OnAddonLoaded(arg1)
    else
      local presetName = BH:DetectSituation()
      if presetName then 
	if     event == "PLAYER_GAINS_VEHICLE_DATA"  then 
	  self:OnPlayerGainsVehicleData()
	elseif event == "VEHICLE_PASSENGERS_CHANGED" then 
	  self:OnVehiclePassengersChanged(presetName)
	end
      end
    end
end

function BH:PostMessage(msg, channel)
    if channel== nil then channel = "local" end
    
    if channel == "local" then
	print(msg);
    else
	local chatType = nil
	
	if channel == "say" then
	    chatType = "SAY"
	elseif channel == "yell" then
	    chatType = "YELL"
	elseif IsInGroup() then
	    if channel == "instance" then
		chatType = "INSTANCE_CHAT"
	    elseif channel == "party" then
		chatType = "PARTY"
	    elseif channel == "raid" then
		chatType = "RAID"
	    elseif channel == "battleground" then
		chatType = "BATTLEGROUND"
	    elseif ((channel == "auto") or (channel == "auto say")) then
		local inInstance, instanceType = IsInInstance()
		if inInstance then
		    if instanceType == "pvp" then
			chatType = "BATTEGROUND"
		    else
			chatType = "INSTANCE_CHAT"
		    end
		else 
		    if IsInRaid() then
			chatType = "RAID"
		    else 
			chatType = "PARTY"
		    end
		end
	    end
	elseif channel == "auto say" then
	    chatType = "SAY"
	end
	
	if chatType then
	    SendChatMessage(msg, chatType)
	end
    end
end
